# -*- coding: utf-8 -*-
import os

from git import Repo
from git import InvalidGitRepositoryError as _InvalidGitRepositoryError

from .domain.config import DeploymentRepoNotFound, TFVarsRepoNotFound
from .domain.config import MissingDockerCredentials, MissingDockerOrg
from .domain.config import EditorNotSpecified
from .domain.config import InvalidRepositoryPath
from .domain.vendor import InvalidGitRepo

config = {
    'TF_DEPLOYMENT_PATH': os.environ.get('TF_DEPLOYMENT_PATH'),
    'TF_DEPLOYMENT_REPO': None,
    'TF_VARS_STAGING_PATH': os.environ.get('TF_VARS_STAGING_PATH'),
    'TF_VARS_STAGING_REPO': None,
    'TF_VARS_PRODUCTION_PATH': os.environ.get('TF_VARS_PRODUCTION_PATH'),
    'TF_VARS_PRODUCTION_REPO': None,
    'DOCKER_USERNAME': os.environ.get('DOCKER_USERNAME'),
    'DOCKER_PASSWORD': os.environ.get('DOCKER_PASSWORD'),
    'DOCKER_ORG': os.environ.get('DOCKER_ORG'),
    'EDITOR': os.environ.get('EDITOR', 'vi'),
}


def validate():
    if not config['TF_DEPLOYMENT_PATH']:
        raise DeploymentRepoNotFound
    if not config['TF_VARS_STAGING_PATH']:
        raise TFVarsRepoNotFound('staging')

    if not config['DOCKER_USERNAME']:
        raise MissingDockerCredentials('username')
    if not config['DOCKER_PASSWORD']:
        raise MissingDockerCredentials('password')
    if not config['DOCKER_ORG']:
        raise MissingDockerOrg()

    if not config['EDITOR']:
        raise EditorNotSpecified()
    return None


def post_process():
    map(_build_repo_from_paths, [
        ('TF_DEPLOYMENT_PATH', 'TF_DEPLOYMENT_REPO',),
        ('TF_VARS_STAGING_PATH', 'TF_VARS_STAGING_REPO',),
        ('TF_VARS_PRODUCTION_PATH', 'TF_VARS_PRODUCTION_REPO',),
    ])


def _build_repo_from_paths(keys):
    key, repo_key = keys
    if config[key] is None:
        return None

    path = os.path.expanduser(config[key])
    if not os.path.isdir(path):
        raise InvalidRepositoryPath(path)
    try:
        repo = Repo(path)
        if repo.bare:
            raise _InvalidGitRepositoryError
        config[repo_key] = repo
    except _InvalidGitRepositoryError:
        raise InvalidGitRepo(path)
    return None



# -*- coding: utf-8 -*-
from . import BaseMimironException
from git import InvalidGitRepositoryError as _InvalidGitRepositoryError


class InvalidGitRepo(BaseMimironException, _InvalidGitRepositoryError):
    def __init__(self, path):
        super(self.__class__, self).__init__('"%s" is an invalid git repository ' % path)


class FetchRemoteUnknownNextStep(BaseMimironException):
    pass


class UnexpectedGitError(BaseMimironException):
    pass


class NoChangesEmptyCommit(BaseMimironException):
    pass



# -*- coding: utf-8 -*-
from . import BaseMimironException


class DeploymentRepoNotFound(BaseMimironException):
    def __init__(self):
        message = 'deployment repository not found'
        super(self.__class__, self).__init__(message)


class TFVarsRepoNotFound(BaseMimironException):
    def __init__(self, env):
        message = '"%s" environment repository not specified' % env
        super(self.__class__, self).__init__(message)


class MissingDockerCredentials(BaseMimironException):
    def __init__(self, missing_credential):
        message = 'missing docker credentials "%s"' % missing_credential
        super(self.__class__, self).__init__(message)


class MissingDockerOrg(BaseMimironException):
    def __init__(self):
        message = '$DOCKER_ORG var was not specified'
        super(self.__class__, self).__init__(message)


class InvalidRepositoryPath(BaseMimironException):
    def __init__(self, path):
        super(self.__class__, self).__init__('"%s" directory does not exist' % path)


class EditorNotSpecified(BaseMimironException):
    def __init__(self):
        super(self.__class__, self).__init__('$EDITOR var was not specified')



# -*- coding: utf-8 -*-
from . import Command as _Command


class Deploy(_Command):
    def _run(self):
        """
        determine the environment
        pull the repo we care about
        load the hcl var file as json
        determine the service we want to update
        make a request on dockerhub to pull down tags for environment
        prompt user to select a tag
        search for the associated var to update
        save hcl file
        set the commit as chore(services): bump <service> image tag <id>
        do a set_var and push
        get the new sha of the commit we just made
        do a set_sha and push
        """
        pass



# -*- coding: utf-8 -*-
from ...config import config

from ..environments import is_valid_env
from ..environments import get_env_repo

from ...domain.commands import InvalidEnvironment
from ...domain.commands import EnvironmentNotLinked


class Command(object):
    def __init__(self, conf=config, **kwargs):
        self.kwargs = kwargs
        self.config = conf

        self._validate_and_configure()
        self._before_run()
        self._run()
        self._after_run()

    def _validate_and_configure(self):
        env = self.kwargs.get('env')
        if env is not None and not is_valid_env(env):
            raise InvalidEnvironment(env)

        self.env = env
        self.repo = get_env_repo(self.env)
        if self.env and not self.repo:
            raise EnvironmentNotLinked(env)

    def _run(self):
        raise NotImplementedError

    def _before_run(self):
        pass

    def _after_run(self):
        pass



# -*- coding: utf-8 -*-
from . import Command as _Command
from ...vendor import git_extensions as git
from .. import io


class Sync(_Command):
    def _run(self):
        for repo in [
            self.config['TF_DEPLOYMENT_REPO'],
            self.config['TF_VARS_STAGING_REPO'],
            self.config['TF_VARS_PRODUCTION_REPO'],
        ]:
            if repo is None:
                continue
            git.sync_updates(repo, push=True)
            git.sync_submodule_updates(repo)
        io.ok('repository sync complete')
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import sys

from terminaltables import SingleTable

# @see: http://pueblo.sourceforge.net/doc/manual/ansi_color_codes.html
U_ERROR = '\033[1;31mERR ✖\033[0m'
U_WARNING = '\033[33m[warn]\033[0m'
U_INFO = '\033[34m==>\033[0m'
U_OK = '\033[32mOK ✓\033[0m'


def _to_utf8(message):
    try:
        return message.encode('utf-8')
    except UnicodeDecodeError:
        return message


def _print_message(stream, *components):
    message = ' '.join(map(unicode, components))
    return print(message, file=stream)


def err(message, exit_=False):
    _print_message(sys.stderr, U_ERROR, message)
    if exit_:
        exit(0)


def warn(message):
    _print_message(sys.stderr, U_WARNING, message)


def info(message):
    _print_message(sys.stdout, U_INFO, message)


def ok(message):
    _print_message(sys.stdout, U_OK, message)


def print_table(rows, title):
    print(SingleTable(rows, title).table)


def collect_input(prompt, selection):
    """Prompts the user to select an option in the `selection` list."""
    message = _to_utf8('%s %s ' % (U_INFO, prompt))
    while True:
        id_ = raw_input(message)
        if id_ == 'q':
            return None
        try:
            id_ = int(id_) - 1
            if id_ < 0:
                continue
            return selection[id_]
        except (ValueError, IndexError):
            continue
    return None



# -*- coding: utf-8 -*-
import requests
import json


class DockerHubAuthentication(object):
    def __init__(self, username, password, org):
        self.username = username
        self.password = password
        self.org = org

        self._token = None

    def generate_token(self):
        payload = json.dumps({
            'username': self.username, 'password': self.password,
        })
        headers = {
            'Content-Type': 'application/json',
        }
        endpoint = 'https://hub.docker.com/v2/users/login/'

        response = requests.post(endpoint, data=payload, headers=headers)
        return response.json()['token'] if response.status_code == 200 else None

    @property
    def token(self):
        if not self._token:
            self._token = self.generate_token()
        return self._token

    @token.setter
    def token(self, new_token):
        self._token = new_token


def _api_request(endpoint, method, auth):
    token = auth.token
    if token is None:
        return None
    response = method(endpoint, headers={'Authorization': 'JWT %s' % token})
    return response.json() if response.status_code == 200 else None


def list_image_repos(auth, page_size=100):
    endpoint = 'https://hub.docker.com/v2/repositories/%s/?page_size=%s' % (
        auth.org, page_size,
    )
    response = _api_request(endpoint, requests.get, auth)
    return response['results'] if response is not None else response


def list_image_tags(image_name, auth, page_size=100):
    endpoint = 'https://hub.docker.com/v2/repositories/%s/%s/tags/?page_size=%s' % (
        auth.org, image_name, page_size,
    )
    response = _api_request(endpoint, requests.get, auth)
    return response['results'] if response is not None else response



# -*- coding: utf-8 -*-
import os

from . import Command as _Command
from .. import io
from ..environments import get_env_repo_name
from ...vendor import git_extensions as git


class SetSha(_Command):
    def _validate_and_configure(self):
        super(self.__class__, self)._validate_and_configure()

        self.deploy_repo = self.config['TF_DEPLOYMENT_REPO']
        self.should_push = self.kwargs['should_push']

    def _prompt_recent_commits(self, commits):
        repo_name = os.path.split(self.repo.working_dir)[-1]
        io.info('listing recent commits for "%s"' % repo_name)

        table_data = [
            ['id', 'commit', 'committed at', 'author (name)'],
        ]
        for i, c in enumerate(commits, 1):
            sha = c.name_rev.split(' ')[0][:10]
            authored_at = c.authored_datetime.strftime('%a %d %b, %I:%M%p')
            table_data.append([i, sha, authored_at, c.author.name])

        io.print_table(table_data, 'recent commits')
        return io.collect_input('select the commit sha you want to update [q]:', commits)

    def _run(self):
        git.sync_updates(self.repo)
        git.sync_updates(self.deploy_repo)

        recent_commits = git.get_recent_commits(self.repo)
        commit = self._prompt_recent_commits(recent_commits)
        if commit is None:
            return None

        submodule_name = get_env_repo_name(self.env)
        commit_sha = commit.name_rev.split(' ')[0]
        has_changes = git.update_submodule(self.deploy_repo, submodule_name, commit_sha)

        if not has_changes:
            return None

        message = 'chore(tfvars): bump %s.tfvars %s' % (self.env, commit_sha[:10])
        if self.should_push:
            git.commit_and_push(self.deploy_repo, message)
        else:
            git.commit_changes(self.deploy_repo, message)
            io.warn('commit to tfvars was NOT pushed to remote!')
            io.warn("it's your responsibility to bundle more changes and explicitly push")



# -*- coding: utf-8 -*-
from . import BaseMimironException


class InvalidEnvironment(BaseMimironException):
    def __init__(self, env):
        message = '"%s" environment is invalid' % env
        super(self.__class__, self).__init__(message)


class EnvironmentNotLinked(BaseMimironException):
    def __init__(self, env):
        message = '"%s" environment is not linked (ref `env`)' % env
        super(self.__class__, self).__init__(message)



# -*- coding: utf-8 -*-
from functools import wraps

import os

from git import Repo
from git.exc import GitCommandError

from ..domain.vendor import UnexpectedGitError
from ..domain.vendor import FetchRemoteUnknownNextStep
from ..domain.vendor import NoChangesEmptyCommit
from ..core import io


def git_failure(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except GitCommandError as e:
            raise UnexpectedGitError(e)
    return wrapper


@git_failure
def get_ahead_behind_count(repo):
    branch = repo.active_branch

    commits_behind = repo.iter_commits('%s..origin/%s' % (branch, branch))
    commits_behind = sum(1 for _ in commits_behind)

    commits_ahead = repo.iter_commits('origin/%s..%s' % (branch, branch))
    commits_ahead = sum(1 for _ in commits_ahead)

    return commits_ahead, commits_behind


@git_failure
def sync_updates(repo, push=False):
    """Updates the local `repo` with origin, pulling and push when necessary.

    True is returned when any changes were made (aside fetch), otherwise False.
    """
    ref = repo.remotes.origin.refs[0].name
    repo_dir = repo.working_dir
    branch = repo.active_branch

    io.info('fetching remote for "%s"@:%s' % (os.path.split(repo_dir)[-1], branch))
    repo.remotes.origin.fetch()

    ahead, behind = get_ahead_behind_count(repo)
    is_dirty = repo.is_dirty()

    if not ahead and not behind and not is_dirty:
        return False

    io.info('(found) [ahead: %s, behind: %s] [dirty: %s]' % (ahead, behind, is_dirty))

    # possible merge conflicts
    if (ahead and behind) or (behind and is_dirty):
        raise FetchRemoteUnknownNextStep('possible merge conflict. please manually resolve')

    # we're ahead so let's push these changes up
    if ahead and push:
        io.warn('ahead, pushing local changes to %s' % ref)
        repo.remotes.origin.push()

    # we're behind so let's pull changes down
    if behind:
        io.warn('behind, pulling changes from %s' % ref)
        repo.remotes.origin.pull()
    return True


@git_failure
def sync_submodule_updates(repo):
    has_updated = False
    for sm in repo.submodules:
        sm.update()
        has_updated = True
    return has_updated


@git_failure
def update_submodule(repo, submodule_name, commit):
    io.info('updating submodule "%s:%s"' % (submodule_name, commit))
    target_sm = None
    for sm in repo.submodules:
        sm.update()
        sm_repo_name = os.path.splitext(os.path.split(sm.url)[-1])[0]
        if sm_repo_name == submodule_name:
            target_sm = sm
            break
    if target_sm is None:
        return io.err("couldn't find submodule to sync updates")

    current_commit = target_sm.branch.commit.name_rev
    current_commit = current_commit.split(' ')[0]

    # selected commit same as current, give up
    if current_commit == commit:
        io.info('"%s" is the same as HEAD')
        return False

    sm_abs_path = os.path.join(repo.working_dir, target_sm.path)
    sm_repo = Repo(sm_abs_path)
    sm_repo.git.checkout(commit)

    return True


@git_failure
def commit_changes(repo, commit_message):
    if not repo.is_dirty():
        return False

    repo.git.add(u=True)
    commit = repo.index.commit(commit_message)
    io.info('created commit with message: "%s"' % commit_message)
    io.info('commit id: %s' % commit.name_rev)
    return True


@git_failure
def push_commits(repo):
    io.info('pushing changes to %s' % repo.remotes.origin.refs[0].name)
    repo.remotes.origin.push()
    io.ok('successfully pushed changes to %s' % repo.remotes.origin.refs[0].name)


@git_failure
def commit_and_push(repo, commit_message):
    if not commit_changes(repo, commit_message):
        raise NoChangesEmptyCommit('"%s" has nothing to commit' % repo.working_dir)
    push_commits(repo)


@git_failure
def get_recent_commits(repo, limit=10):
    return list(repo.iter_commits(max_count=limit))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mimiron.py

usage:
    mim deploy <service> <env> [<artifact>]
    mim set-sha <env> [--no-push]
    mim set-var <env>
    mim sync
    mim edit <env>

commands:
    deploy        update ami/sha and auto-deploy after update
    set-sha       update tfvars commit sha in deployments and push to remote
    set-var       commit and pushes tfvars for <env> based on changes found
    sync          calls git fetch on all associated git repos
    edit          opens the tfvar config in the default editor

arguments:
    <artifact>    the deployment artifact we are pushing (e.g. Docker image/AMI)
    <service>     the application/microservice we're targeting
    <env>         the environment we want to change

options:
    --no-push     make local changes without pushing to remote

    -h --help     shows this
    -v --version  shows version
"""
from __future__ import print_function

from . import __version__
from docopt import docopt

import config
from .core import io
from .core.commands import deploy, sync, edit
from .core.commands import set_sha, set_var
from .domain import BaseMimironException


def _parse_user_input(args):
    env = args['<env>']

    if args['deploy']:
        return deploy.Deploy(
            env=env, service=args['<service>'], artifact=args['<artifact>']
        )
    if args['set-sha']:
        should_push = not args['--no-push']
        return set_sha.SetSha(env=env, should_push=should_push)
    if args['set-var']:
        return set_var.SetVar(env=env)
    if args['sync']:
        return sync.Sync()
    if args['edit']:
        return edit.Edit(env=env)
    io.err('encountered unexpected mim command')


def main():
    args = docopt(__doc__, version=__version__)
    try:
        config.validate()
        config.post_process()

        _parse_user_input(args)
    except KeyboardInterrupt:
        pass
    except BaseMimironException as e:
        io.err(e)
    exit(0)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import os
import subprocess

from . import Command as _Command
from .. import io


class Edit(_Command):
    def _run(self):
        io.info('preparing to open config for "%s"' % self.env)

        config_var_path = os.path.join(self.repo.working_dir, 'variables.tfvars')
        if not os.path.isfile(config_var_path):
            io.err('failed to open "%s" (not a file)' % config_var_path)
        else:
            io.info('opening "%s"' % config_var_path)
            subprocess.call([self.config['EDITOR'], config_var_path])
        io.info('closed "%s" config, finished editing' % self.env)



# -*- coding: utf-8 -*-
import os
from ..config import config

STAGING = 'staging'
PRODUCTION = 'production'


def is_valid_env(env):
    return env.lower() in [PRODUCTION, STAGING]  # only 2 for now.


def get_env_repo(env):
    if env == PRODUCTION:
        return config['TF_VARS_PRODUCTION_REPO']
    if env == STAGING:
        return config['TF_VARS_STAGING_REPO']
    return None


def get_env_repo_name(env):
    repo_name = None
    if env == PRODUCTION:
        repo_name = config['TF_VARS_PRODUCTION_PATH']
    if env == STAGING:
        repo_name = config['TF_VARS_STAGING_PATH']
    return os.path.split(repo_name)[-1] if repo_name else repo_name



# -*- coding: utf-8 -*-
from . import Command as _Command
from ...vendor import git_extensions as git


class SetVar(_Command):
    def _run(self):
        has_changes = git.sync_updates(self.repo)
        if has_changes:
            message = 'chore(variables): updated variables.tfvars file'
            git.commit_and_push(self.repo, message)
        return None
